@@ -1,10 +1,12 @@
+import yaml
 import ujson
 import urllib
 import os
+import sys
 import httplib
 import time
-import math
 import urllib2
+import fnmatch
 
 import jsonpatch
 from boto.s3.connection import S3Connection, OrdinaryCallingFormat
@@ -18,28 +20,39 @@ from . import __version__
 # Item class
 #_________________________________________________________________________________________
 class Item(object):
-    """This class represents an archive.org item.
-    You can use this class to access item metadata:
+    """This class represents an archive.org item. You can use this 
+    class to access item metadata:
         >>> import internetarchive
         >>> item = internetarchive.Item('stairs')
         >>> print item.metadata
 
-    This class also uses IA's S3-like interface to upload files to an item. You need to
-    supply your IAS3 credentials in environment variables in order to upload. You can
-    retrieve S3 keys from https://archive.org/account/s3.php
-        >>> import os;
-        >>> os.environ['AWS_ACCESS_KEY_ID']='x'; os.environ['AWS_SECRET_ACCESS_KEY']='y'
-        >>> item.upload('myfile')
+    This class also uses IA's S3-like interface to upload files to an 
+    item. You need to supply your IAS3 credentials in environment 
+    variables in order to upload. You can retrieve S3 keys from:
+    https://archive.org/account/s3.php
+        >>> import os
+        >>> os.environ['AWS_ACCESS_KEY_ID']='$access_key'
+        >>> os.environ['AWS_SECRET_ACCESS_KEY']='$secret_key'
+        >>> item.upload('myfile.tar')
         True
     """
 
     # init()
     #_____________________________________________________________________________________
-    def __init__(self, identifier, metadata_timeout=None):
+    def __init__(self, identifier, metadata_timeout=None, host=None, s3_keys=None,
+                 config=None, config_file=None):
+        self.config = config
+        self.config_file = config_file
+        self.s3_keys = s3_keys
         self.identifier = identifier
-        self.details_url = 'https://archive.org/details/{0}'.format(identifier)
-        self.download_url = 'https://archive.org/download/{0}'.format(identifier)
-        self.metadata_url = 'https://archive.org/metadata/{0}'.format(identifier)
+        if host:
+            _url_prefix = 'https://{0}.'.format(host)
+        else:
+            _url_prefix = 'https://'
+        self.details_url = '{0}archive.org/details/{1}'.format(_url_prefix, identifier)
+        # TODO: https is not working with gevent
+        self.download_url = '{0}archive.org/download/{1}'.format(_url_prefix, identifier)
+        self.metadata_url = '{0}archive.org/metadata/{1}'.format(_url_prefix, identifier)
         self.metadata_timeout = metadata_timeout
         self._s3_conn = None
         self._bucket = None
@@ -50,6 +63,28 @@ class Item(object):
             self.exists = True
 
 
+    # _configure()
+    #_____________________________________________________________________________________
+    def _configure(self):
+        if not self.s3_keys:
+            config_file = os.path.join(os.environ['HOME'], '.config', 
+                                       'internetarchive.yml')
+            if os.path.exists(config_file):
+                self.config_file = config_file
+            else:
+                config_file = os.path.join(os.environ['HOME'], '.internetarchive.yml')
+                self.config_file = config_file
+            if self.config_file:
+                self.config = yaml.load(open(self.config_file))
+                if self.config:
+                    self.s3_keys = self.config.get('s3') 
+            if not self.s3_keys:
+                self.s3_keys = dict(
+                        access_key = os.environ['AWS_ACCESS_KEY_ID'],
+                        secret_key = os.environ['AWS_SECRET_ACCESS_KEY'],
+                )
+
+
     # _get_item_metadata()
     #_____________________________________________________________________________________
     def _get_item_metadata(self):
@@ -80,30 +115,67 @@ class Item(object):
 
     # download()
     #_____________________________________________________________________________________
-    def download(self, formats=None):
+    def download(self, source=None, formats=None, concurrent=False, glob_pattern=None, 
+                 ignore_existing=False):
         """Download the entire item into the current working directory"""
-        for f in self.files():
+        if concurrent:
+            try:
+                from gevent import monkey
+                monkey.patch_socket()
+                from gevent.pool import Pool
+                pool = Pool()
+            except ImportError:
+                raise ImportError(
+                """No module named gevent
+
+                Downloading files concurrently requires the gevent neworking library.
+                gevent and all of it's dependencies can be installed with pip:
+                
+                \tpip install cython git+git://github.com/surfly/gevent.git@1.0rc2#egg=gevent
+
+                """)
+
+        files = self.files()
+        if source:
+            if type(source) == str:
+                source = [source]
+            files = [f for f in files if f.source in source]
+        if formats:
             if type(formats) == str:
                 formats = [formats]
-            if formats is not None and f.format not in formats:
-                continue
-            print '  downloading', f.name
-            path = os.path.join(self.identifier, f.name)
-            parent_dir = os.path.dirname(path)
-            f.download(path)
+            files = [f for f in files if f.format in formats]
+        if glob_pattern:
+            files = [f for f in files if fnmatch.fnmatch(f.name, glob_pattern)]
+
+        for f in files:
+            fname = f.name.encode('utf-8')
+            path = os.path.join(self.identifier, fname)
+            sys.stdout.write('downloading: {0}\n'.format(fname))
+            if concurrent:
+                pool.spawn(f.download, path, ignore_existing=ignore_existing)
+            else:
+                f.download(path, ignore_existing=ignore_existing)
+        if concurrent:
+            pool.join()
 
 
     # modify_metadata()
     #_____________________________________________________________________________________
     def modify_metadata(self, metadata, target='metadata'):
-        """function for modifying the metadata of an existing item on archive.org.
-        Note: The Metadata Write API does not yet comply with the latest Json-Patch
-        standard. It currently complies with version 02:
+        """function for modifying the metadata of an existing item on 
+        archive.org.
+
+        Note: The Metadata Write API does not yet comply with the 
+        latest Json-Patch standard. It currently complies with version 
+        02:
 
         https://tools.ietf.org/html/draft-ietf-appsawg-json-patch-02
 
-        :param metadata: Dictionary. Metadata used to update the item.
-        :param target: (optional) String. Metadata target to update.
+        :type metadata: dict
+        :param metadata: Metadata used to update the item.
+
+        :type target: str
+        :param target: (optional) Set the metadata target to update.
 
         Usage:
 
@@ -113,8 +185,8 @@ class Item(object):
         >>> item.modify_metadata(md)
 
         """
-        access_key = os.environ['AWS_ACCESS_KEY_ID']
-        secret_key = os.environ['AWS_SECRET_ACCESS_KEY']
+        if not self.s3_keys:
+            self._configure()
         src = self.metadata.get(target, {})
         dest = dict((src.items() + metadata.items()))
 
@@ -135,8 +207,8 @@ class Item(object):
         data = {
             '-patch': ujson.dumps(patch),
             '-target': target,
-            'access': access_key,
-            'secret': secret_key,
+            'access': self.s3_keys['access_key'],
+            'secret': self.s3_keys['secret_key'],
         }
 
         host = 'archive.org'
@@ -162,7 +234,11 @@ class Item(object):
     #_____________________________________________________________________________________
     def _get_s3_conn(self):
         if self._s3_conn is None:
-            self._s3_conn = S3Connection(host='s3.us.archive.org', is_secure=False,
+            if not self.s3_keys:
+                self._configure()
+            self._s3_conn = S3Connection(self.s3_keys['access_key'], 
+                                         self.s3_keys['secret_key'],
+                                         host='s3.us.archive.org', 
                                          calling_format=OrdinaryCallingFormat())
         return self._s3_conn
 
@@ -170,7 +246,7 @@ class Item(object):
     # _get_s3_bucket()
     #_____________________________________________________________________________________
     def _get_s3_bucket(self, conn, headers={}, ignore_bucket=False):
-        if ignore_bucket is True:
+        if ignore_bucket:
             headers['x-archive-ignore-preexisting-bucket'] = 1
             self._bucket = None
         else:
@@ -215,29 +291,55 @@ class Item(object):
     def upload_file(self, _file, remote_name=None, metadata={}, headers={}, derive=True,
                     ignore_bucket=False, multipart=False, bytes_per_chunk=16777216,
                     debug=False):
-        """Upload a single file to an item. The item will be created if it does not exist.
+        """Upload a single file to an item. The item will be created 
+        if it does not exist.
 
-        :param _file: String or File. The filepath or file-like object to be uploaded.
-        :param remote_name: (optional) String. Sets the remote filename.
-        :param metadata: (optional) Dictionary. Metadata used to create a new item.
-        :param headers: (optional) Dictionary. Add additional IA-S3 headers to request.
-        :param derive: (optional) Boolean. Set to False to prevent an item from being derived after upload.
-        :param multipart: (optional) Boolean. Set to True to upload files in parts. Useful when uploading large files.
-        :param ignore_bucket: (optional) Boolean. Set to True to ignore and clobber existing files and metadata.
-        :param debug: (optional) Boolean. Set to True to print headers to stdout -- don't upload anything.
-        :param bytes_per_chunk: (optional) Integer. Used to determine the chunk size when using multipart upload.
+        :type _file: str or file
+        :param _file: The filepath or file-like object to be uploaded.
+
+        :type remote_name: str
+        :param remote_name: (optional) Sets the remote filename.
+
+        :type metadata: dict
+        :param metadata: (optional) Metadata used to create a new item.
+
+        :type headers: dict
+        :param headers: (optional) Add additional IA-S3 headers to 
+                        request.
+
+        :type derive: bool
+        :param derive: (optional) Set to False to prevent an item from 
+                       being derived after upload.
+
+        :type multipart: bool
+        :param multipart: (optional) Set to True to upload files in 
+                          parts. Useful when uploading large files.
+
+        :type ignore_bucket: bool
+        :param ignore_bucket: (optional) Set to True to ignore and 
+                              clobber existing files and metadata.
+
+        :type debug: bool
+        :param debug: (optional) Set to True to print headers to stdout,
+                      and exit without sending the upload request.
+
+        :type bytes_per_chunk: int
+        :param bytes_per_chunk: (optional) Used to determine the chunk 
+                                size when using multipart upload.
 
         Usage::
 
             >>> import internetarchive
             >>> item = internetarchive.Item('identifier')
-            >>> item.upload_file('/path/to/image.jpg', remote_name='photos/image1.jpg')
+            >>> item.upload_file('/path/to/image.jpg', 
+            ...                  remote_name='photos/image1.jpg')
             True
-
         """
 
         headers = self._get_s3_headers(headers, metadata)
-        headers['x-archive-size-hint'] = os.stat(_file).st_size
+        header_names = [header_name.lower() for header_name in headers.keys()]
+        if 'x-archive-size-hint' not in header_names:
+            headers['x-archive-size-hint'] = os.stat(_file).st_size
         scanner = 'Internet Archive Python library {0}'.format(__version__)
         headers['x-archive-meta-scanner'] = scanner
 
@@ -249,14 +351,14 @@ class Item(object):
         conn = self._get_s3_conn()
         bucket = self._get_s3_bucket(conn, headers, ignore_bucket=ignore_bucket)
 
-        if derive is False:
+        if not derive:
             headers['x-archive-queue-derive'] =  0
 
         # Don't clobber existing files unless ignore_bucket is True.
-        if bucket.get_key(remote_name) and ignore_bucket is False:
+        if bucket.get_key(remote_name) and not ignore_bucket:
             return True
 
-        if multipart is False:
+        if not multipart:
             k = boto.s3.key.Key(bucket)
             k.name = remote_name
             k.set_contents_from_file(_file, headers=headers)
@@ -270,23 +372,21 @@ class Item(object):
                 mp.upload_part_from_file(part_fp, part_num=part)
                 part += 1
             mp.complete_upload()
-            #    mp.cancel_upload()
         return True
 
 
     # upload()
     #_____________________________________________________________________________________
-    def upload(self, files, metadata={}, headers={}, derive=True, ignore_bucket=False, 
-               multipart=False, bytes_per_chunk=16777216, debug=False):
-        """Upload files to an item. The item will be created if it does not exist.
+    def upload(self, files, **kwargs):
+        """Upload files to an item. The item will be created if it 
+        does not exist.
 
-        :param files: List. The filepaths or file-like objects to upload.
-        :param metadata: (optional) Dictionary. Metadata used to create a new item.
-        :param derive: (optional) Boolean. Set to False to prevent an item from being derived after upload.
-        :param ignore_bucket: (optional) Boolean. Set to True to ignore and clobber existing files and metadata.
-        :param multipart: (optional) Boolean. Set to True to upload files in parts. Useful when uploading large files.
-        :param bytes_per_chunk: (optional) Integer. Used to determine the chunk size when using multipart upload.
-        :param debug: (optional) Boolean. Set to True to print headers to stdout -- don't upload anything.
+        :type files: list
+        :param files: The filepaths or file-like objects to upload.
+
+        :type kwargs: dict
+        :param kwargs: The keyword arguments from the call to 
+                       upload_file().
 
         Usage::
 
@@ -295,19 +395,16 @@ class Item(object):
             >>> md = dict(mediatype='image', creator='Jake Johnson')
             >>> item.upload('/path/to/image.jpg', md, derive=False)
             True
-
         """
 
-        if debug is True:
-            return self._get_s3_headers(headers, metadata)
+        if kwargs.get('debug'):
+            return self._get_s3_headers(kwargs.get('headers', {}), 
+                                        kwargs.get('metadata', {}))
         if type(files) != list:
             files = [files]
         for _file in files:
-            upload_status = self.upload_file(_file, remote_name=None, metadata=metadata, 
-                                                    headers=headers, derive=derive, 
-                                                    ignore_bucket=ignore_bucket,
-                                                    multipart=multipart) 
-            if upload_status is True:
+            upload_status = self.upload_file(_file, **kwargs)
+            if upload_status:
                 continue
             else:
                 return False
@@ -338,27 +435,31 @@ class File(object):
 
     # download()
     #_____________________________________________________________________________________
-    def download(self, file_path=None):
+    def download(self, file_path=None, ignore_existing=False):
         if file_path is None:
             file_path = self.name
 
-        if os.path.exists(file_path):
+        if os.path.exists(file_path) and not ignore_existing:
             raise IOError('File already exists: {0}'.format(file_path))
 
         parent_dir = os.path.dirname(file_path)
         if parent_dir != '' and not os.path.exists(parent_dir):
             os.makedirs(parent_dir)
 
-        url = 'https://archive.org/download/{0}/{1}'.format(self.item.identifier,
-                                                            self.name)
+        fname = self.name.encode('utf-8')
+        url = '{0}/{1}'.format(self.item.download_url, fname)
         urllib.urlretrieve(url, file_path)
 
 
 # Search class
 #_________________________________________________________________________________________
 class Search(object):
-    """This class represents an archive.org item search. You can use this class to search
-    for archive.org items using the advanced search engine:
+    """This class represents an archive.org item search. You can use 
+    this class to search for archive.org items using the advanced 
+    search engine.
+
+    Usage::
+
         >>> import internetarchive
         >>> search = internetarchive.Search('(uploader:jake@archive.org)')
         >>> for result in search.results:
@@ -368,7 +469,7 @@ class Search(object):
     # init()
     #_____________________________________________________________________________________
     def __init__(self, query, fields=['identifier'], params={}):
-        self._base_url = 'https://archive.org/advancedsearch.php'
+        self._base_url = 'http://archive.org/advancedsearch.php'
         self.query = query
         self.params = dict(dict(
                 q = self.query,
@@ -412,7 +513,96 @@ class Search(object):
                 yield doc
 
 
-# Catalog class
+# Mine class
+#_________________________________________________________________________________________
+class Mine(object):
+    """
+    Usage::
+        >>> miner = internetarchive.Mine('itemlist.txt', workers=50)
+        >>> for md in miner:
+        ...     print md
+    """
+    # __init__()
+    #_____________________________________________________________________________________
+    def __init__(self, identifiers, workers=20):
+        try:
+            from gevent import monkey, queue
+            monkey.patch_all()
+        except ImportError:
+            raise ImportError(
+            """No module named gevent
+
+            This feature requires the gevent neworking library.  gevent 
+            and all of it's dependencies can be installed with pip:
+            
+            \tpip install cython git+git://github.com/surfly/gevent.git@1.0rc2#egg=gevent
+
+            """)
+
+        self.hosts = None
+        self.skips = []
+        self.queue = queue
+        self.workers = workers
+        self.done_queueing_input = False
+        self.queued_count = 0
+        self.identifiers = identifiers
+        self.input_queue = self.queue.JoinableQueue(1000)
+        self.json_queue = self.queue.Queue(1000)
+
+
+    # _metadata_getter()
+    #_____________________________________________________________________________________
+    def _metadata_getter(self):
+        import random
+        while True:
+            i, identifier = self.input_queue.get()
+            if self.hosts:
+                host = self.hosts[random.randrange(len(self.hosts))]
+                while host in self.skips:
+                    host = hosts[random.randrange(len(self.hosts))]
+            else:
+                host = None
+            try:
+                item = Item(identifier, host=host)
+                self.json_queue.put((i, item))
+            except:
+                if host:
+                    sys.stderr.write('host failed: {0}\n'.format(host))
+                    self.skips.append(host)
+                self.input_queue.put((i, identifier))
+            finally:
+                self.input_queue.task_done()
+
+
+    # _queue_input()
+    #_____________________________________________________________________________________
+    def _queue_input(self):
+        for i, identifier in enumerate(self.identifiers):
+            self.input_queue.put((i, identifier))
+            self.queued_count += 1
+        self.done_queueing_input = True
+
+
+    # items()
+    #_____________________________________________________________________________________
+    def items(self):
+        import gevent
+        gevent.spawn(self._queue_input)
+        for i in range(self.workers):
+            gevent.spawn(self._metadata_getter)
+
+        def metadata_iterator_helper():
+            got_count = 0
+            while True:
+                if self.done_queueing_input and got_count == self.queued_count:
+                    break
+                yield self.json_queue.get()
+                got_count += 1
+
+        return metadata_iterator_helper()
+
+
+
 #_________________________________________________________________________________________
 class Catalog(object):
     GREEN = 0
@@ -422,7 +612,10 @@ class Catalog(object):
 
     # init()
     #_____________________________________________________________________________________
-    def __init__(self, params=None):
+    def __init__(self, params=None, cookies=None):
+        self.cookies = cookies
+        if not self.cookies:
+            self.config = self._configure()
         url = 'http://archive.org/catalog.php'
 
         if params is None:
@@ -434,9 +627,9 @@ class Catalog(object):
         params['callback'] = 'foo'
         params = urllib.urlencode(params)
 
-        ia_cookies = ('logged-in-sig={LOGGED_IN_SIG}; '
-                      'logged-in-user={LOGGED_IN_USER}; '
-                      'verbose=1'.format(**os.environ))
+        ia_cookies = ('logged-in-sig={logged-in-sig}; '
+                      'logged-in-user={logged-in-user}; '
+                      'verbose=1'.format(**self.cookies))
 
         opener = urllib2.build_opener()
         opener.addheaders.append(('Cookie', ia_cookies))
@@ -449,25 +642,69 @@ class Catalog(object):
         tasks_json = ujson.loads(json_str)
         self.tasks = [CatalogTask(t) for t in tasks_json]
         
+
+    # _configure()
+    #_____________________________________________________________________________________
+    def _configure(self):
+        config_file = os.path.join(os.environ['HOME'], '.config', 
+                                   'internetarchive.yml')
+        if os.path.exists(config_file):
+            self.config_file = config_file
+        else:
+            config_file = os.path.join(os.environ['HOME'], '.internetarchive.yml')
+            self.config_file = config_file
+        if self.config_file:
+            self.config = yaml.load(open(self.config_file))
+            if self.config:
+                self.cookies = self.config.get('cookies') 
+        if not self.cookies:
+            self.cookies = {
+                    'logged-in-user': os.environ['LOGGED_IN_USER'],
+                    'logged-in-sig': os.environ['LOGGED_IN_sig'],
+            }
+
+
+    # filter_tasks()
+    #_____________________________________________________________________________________
     def filter_tasks(self, pred):
         return [t for t in self.tasks if pred(t)]
 
+
+    # tasks_by_type()
+    #_____________________________________________________________________________________
     def tasks_by_type(self, row_type):
         return self.filter_tasks(lambda t: t.row_type == row_type)
 
+    # green_rows()
+    #_____________________________________________________________________________________
     @property
     def green_rows(self):
         return self.tasks_by_type(self.GREEN)
+
+
+    # blue_rows()
+    #_____________________________________________________________________________________
     @property
     def blue_rows(self):
         return self.tasks_by_type(self.BLUE)
+
+
+    # red_rows()
+    #_____________________________________________________________________________________
     @property
     def red_rows(self):
         return self.tasks_by_type(self.RED)
+
+
+    # brown_rows()
+    #_____________________________________________________________________________________
     @property
     def brown_rows(self):
         return self.tasks_by_type(self.BROWN)
 
+
+# CatalogTask class
+#_________________________________________________________________________________________
 class CatalogTask(object):
     """represents catalog task.
     """
@@ -504,5 +741,3 @@ class CatalogTask(object):
             raise ValueError, 'task_id is None'
         url = 'http://catalogd.archive.org/log/{0}'.format(self.task_id)
         return urllib2.urlopen(url)
-
-        
